// Types which are only used when compiling CDL files
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.


#ifndef CDL_H
#define CDL_H

#include <idlc.h>

class Class;
typedef Ref<Class> ClassRef;

extern list<ClassRef> classes;

class Class : public NameSpace {
	list<InterfaceRef> ifaces;

public:
	struct ParamInfo : public RefCountable<ParamInfo> {
		bool copy;
	};

	typedef Ref<ParamInfo> ParamInfoRef;

private:
	// C++ associative arrays are much more of a 
	// pain to use than they should be.

	typedef map<ParamRef, ParamInfoRef> params_map_type;
	typedef params_map_type::value_type params_valtype;
	typedef params_map_type::const_iterator params_iter;

public:
	struct MethodInfo : public RefCountable<MethodInfo> {
		StrListRef implname;
		bool copy_params; // True if at least one parameter
		                  // has the copy attribute.

	private:
		params_map_type params;
	
	public:
		ParamInfo *get_param(Param *p)
		{
			params_iter ret = params.find(p);
			
			if (ret != params.end())
				return (*ret).second;
			
			return NULL;
		}
		
		ParamInfo *add_param(Param *p)
		{
			ParamInfo *pi = get_param(p);
		
			if (!pi) {
				pi = new ParamInfo;
				pi->copy = false;
				
				pair<params_iter, bool> ret = 
					params.insert(params_valtype(p, pi));
	
				assert(ret.second);
			}
		
			return pi;
		}
	};

	typedef Ref<MethodInfo> MethodInfoRef;

private:
	typedef map<MethodRef, MethodInfoRef> method_map_type;
	typedef method_map_type::value_type methods_valtype;
	typedef method_map_type::const_iterator methods_iter;

	method_map_type methods;

public:
	InterfaceRef concrete_iface;

	Class(const String *name) :
	Symbol(name), concrete_iface(new Interface(new String("<anon>")))
	{
		classes.push_back(this);
	}
	
	void add_iface(Interface *iface)
	{
		ifaces.push_back(iface);
		concrete_iface->add_super(iface);
	}
	
	MethodInfo *get_method(Method *m)
	{
		methods_iter ret = methods.find(m);
		
		if (ret != methods.end())
			return (*ret).second;
		
		return NULL;
	}
	
	MethodInfo *add_method(Method *m)
	{
		MethodInfo *mi = get_method(m);

		if (!mi) {
			mi = new MethodInfo;
			mi->copy_params = false;
			
			pair<methods_iter, bool> ret = 
				methods.insert(methods_valtype(m, mi));

			assert(ret.second);
		}
		
		return mi;
	}

	typedef list<InterfaceRef>::const_iterator ifaces_iterator;

	ifaces_iterator ifaces_begin()
	{
		return ifaces.begin();
	}

	ifaces_iterator ifaces_end()
	{
		return ifaces.end();
	}
	
	// FIXME: check for duplicate method implementation names.
	void finalize()
	{
		concrete_iface->finalize_class_iface();
	}
};

#endif
