#ifndef _CPP_ORB_H
#define _CPP_ORB_H

#include <stdint.h>
#include <stddef.h>
#include <string.h>
#include <limits.h>

#include <lowlevel/barriers.h>
#include <lowlevel/bitops.h>

#include <util/array.h>

namespace System {
	struct _i_Object;

	namespace Mem {
		struct Region;
	}

	namespace RunTime {
		using namespace ::Util::Arrays;
		
		class ORBMM {
			static void *priv;
			
		public:
			typedef ::System::Mem::Region Region;
			
			static void *alloc(size_t size, int refs = 1);
			static void retain(void *ptr, int refs = 1);
			static void release(void *ptr, int refs = 1);
			static void add_region(Region region, bool unmap_orig, int refs = 1);
		};
		
		struct IFaceInfo;
		typedef uint32_t ID;

		struct IFaceTable {
			const IFaceInfo *info;
			const ptrdiff_t offset;
		};
		
		union GUID {
			unsigned char c[16];
			unsigned long l[];
		};

		uintptr_t downcast(::System::_i_Object *obj, const GUID *new_guid);

		// FIXME: use above typedef
		static inline bool guids_equal(const unsigned long *guid1,
		                               const unsigned long *guid2)
		{
			return (guid1[0] == guid2[0] && 
			        guid1[1] == guid2[1] &&
			        (sizeof(long) == 8 ||
			         (guid1[2] == guid2[2] &&
			          guid1[3] == guid2[3])));
		}
		
		struct ParamInfoBlock {
			uint32_t buffer_size;
			uint32_t copy_size;
			ID *objlist, *newobj;
			uint32_t objlist_len, newobj_len;
			uint32_t num_segments;

			struct Segment {
				uint8_t *ptr;
				uintptr_t len;
				uintptr_t flags;
				uintptr_t reserved;
				
				enum {
					In = 1,
					Out = 2,
					Inline = 4,
					Copy = 8
				};
			} segments[0];
		};
		
		struct Segment0 {
			enum {
				InvokeMethod = 0,
				GetIFaces = 1,
				ReturnIFaces = 2,
			};
		
			union {
				struct {
					uint8_t opcode;
				};
				
				unsigned long pad;
			};
		};
			
		struct InvokeMethod {
			Segment0 seg0;
			uint32_t iface;
			uint32_t method;
			uint8_t args[0];
		};
		
		union Message {
			Segment0 seg0;
			InvokeMethod invoke;
		};

		struct IFaceInfo {
			const GUID *guid;
			void (*invoke)(Array<ID, ORBMM> objlist,
			               ParamInfoBlock::Segment *segs,
			               int nsegs);
			::System::_i_Object *(*wrapper)(ID id);
		};
		
		struct NewObject {
			uint32_t guid_hash[5]; // SHA-1 hash of Interface GUIDs
			uint32_t id;
			uint32_t reserved[3]; // must be zero
		};

		struct MarshCtx {
			GrowableArray<uint8_t, ORBMM> &buf;
			GrowableArray<ID, ORBMM> &objlist;
			GrowableArray<NewObject, ORBMM> &newobjlist;
			ParamInfoBlock::Segment *segs;
			int nsegs;
		};

		struct VStructInfo {
			// List of GUIDs of the struct and its superstructs,
			// starting with System.VStruct and ending with
			// the concrete struct.
			
			const unsigned long *const *const guids;

			// Length of inheritance chain; 1 for System.VStruct
			
			const int chainlen;

			int (*marshall)(MarshCtx &ctx);
			void (*unmarshall)(MarshCtx &ctx);
		};

		namespace Priv {
			// Return the caller's PC.  It'd be nice if GCC had a builtin for
			// the current PC, so that a simple relocation could be used rather
			// than a function call.  OPT: put this in lowlevel-lib, so that
			// architectures can provide a faster version.
			
			unsigned long get_pc();

			static inline bool in_kernel()
			{
#ifdef _KERNEL
				return true;
#else
				return false;
#endif
			}

		}
	};
}

inline void *operator new(size_t len, ::System::RunTime::ORBMM orbmm,
                          int refs = 1)
{
	return ::System::RunTime::ORBMM::alloc(len, refs);
}

inline void *operator new[](size_t len, ::System::RunTime::ORBMM orbmm,
                            int refs = 1)
{
	return ::System::RunTime::ORBMM::alloc(len, refs);
}

// This is a macro rather than an inline template function so that the
// caller shows up as file/line number in the debugging information rather
// than this header file, and so that a variable argument list can be
// passed to a fixed arg ctor.
//
// To throw an IDL exception of type Foo, do this:
// throw_idl(Foo, args, to, foo);

#ifndef POLINTOS_NO_THROW_IDL
#define throw_idl(T, args...) do { \
	throw T(NULL, NULL, \
	        new(::System::RunTime::ORBMM()) \
	           ::System::Exceptions::NativeCodeExceptionOriginInfo \
	           (::System::RunTime::Priv::get_pc()), \
	        ::System::RunTime::Priv::in_kernel(), ##args); \
} while (0)

#define rethrow_idl(oldex, T, args...) do { \
	throw T(new(::System::RunTime::ORBMM()) typeof(oldex)(oldex), NULL, \
	        new(::System::RunTime::ORBMM()) \
	           ::System::Exceptions::NativeCodeExceptionOriginInfo \
	           (::System::RunTime::Priv::get_pc()), \
	        ::System::RunTime::Priv::in_kernel(), ##args); \
} while (0)
#endif

#endif
